@@ -4,13 +4,16 @@
 
 """ Public XML parser module """
 
-import lxml.etree
 import re
 import os
-import shutil
-import xml2rfc.log
-import requests
 import codecs
+import shutil
+import six
+import time
+import requests
+import lxml.etree
+import xml2rfc.log
+import xml2rfc.utils
 
 try:
     from urllib.parse import urlparse, urljoin
@@ -42,10 +45,11 @@ class CachingResolver(lxml.etree.Resolver):
     """ Custom ENTITY request handler that uses a local cache """
     def __init__(self, cache_path=None, library_dirs=None, source=None,
                  templates_path='templates', verbose=False, quiet=False,
-                 network_locs= [
+                 no_network=False, network_locs= [
+                     'https://xml2rfc.ietf.org/public/rfc/',
+                     'https://xml2rfc.tools.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
-                     'http://xml.resource.org/public/rfc/',
                  ],
                  rfc_number=None):
         self.verbose = verbose
@@ -53,12 +57,20 @@ class CachingResolver(lxml.etree.Resolver):
         self.source = source
         self.library_dirs = library_dirs
         self.templates_path = templates_path
+        self.no_network=no_network
         self.network_locs = network_locs
         self.include = False
         self.rfc_number = rfc_number
+        self.cache_refresh_secs = (60*60*24*14) # 14 days
 
         # Get directory of source
-        self.source_dir = os.path.abspath(os.path.dirname(self.source))
+        if self.source:
+            if isinstance(self.source, six.string_types):
+                self.source_dir = os.path.abspath(os.path.dirname(self.source))
+            else:
+                self.source_dir = os.path.abspath(os.path.dirname(self.source.name))                
+        else:
+            self.source_dir = None
 
         # Determine cache directories to read/write to
         self.read_caches = [ os.path.expanduser(path) for path in xml2rfc.CACHES ]
@@ -161,7 +173,7 @@ class CachingResolver(lxml.etree.Resolver):
               ['/usr/share/xml2rfc'].  On the GUI this can be configured
               manually but has the same initial defaults.
             - NETWORK refers to the online citation library.  On the CLI this
-              is http://xml.resource.org/public/rfc/.  On the GUI this 
+              is http://xml2rfc.ietf.org/public/rfc/.  On the GUI this 
               can be configured manually but has the same initial default.
 
             The caches in read_dirs are consulted in sequence order to find the
@@ -223,6 +235,8 @@ class CachingResolver(lxml.etree.Resolver):
                         result = self.cache(path)
                         if result:
                             break
+                if not result and self.no_network:
+                    xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
                 tried_cache = True
             else:
                 if os.path.dirname(paths[0]):
@@ -247,6 +261,9 @@ class CachingResolver(lxml.etree.Resolver):
                             if result:
                                 break
                         tried_cache = True
+                        if not result and self.no_network:
+                            xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
+
                         # if not result:
                         #     # Document didn't exist, default to source dir
                         #     result = os.path.join(self.source_dir, request)
@@ -277,13 +294,15 @@ class CachingResolver(lxml.etree.Resolver):
                             tried_cache = True
                             if result:
                                 break
+                        if not result and self.no_network:
+                            xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
                     # if not result:
                     #     # Default to source dir
                     #     result = os.path.join(self.source_dir, request)
                     #     attempts.append(result)
 
         # Verify the result -- either raise exception or return it
-        if not os.path.exists(result) and not urlparse(result).netloc:
+        if not result or (not os.path.exists(result) and not urlparse(original).netloc):
             if os.path.isabs(original):
                 xml2rfc.log.warn('A reference was requested with an absolute path, but not found '
                     'in that location.  Removing the path component will cause xml2rfc to look for'
@@ -315,10 +334,14 @@ class CachingResolver(lxml.etree.Resolver):
         for dir in self.read_caches:
             cached_path = os.path.join(dir, xml2rfc.CACHE_PREFIX, basename)
             if os.path.exists(cached_path):
-                xml2rfc.log.note('Resolving ' + typename + '...', url)
-                xml2rfc.log.note('Loaded from cache', cached_path)
-                return cached_path
-        # Not found, save to `write_cache`
+                if os.path.getmtime(cached_path) < (time.time() - self.cache_refresh_secs) and not self.no_network:
+                    xml2rfc.log.note('Cached version at %s too old; will refresh cache for %s %s' % (cached_path, typename, url))
+                    break
+                else:
+                    xml2rfc.log.note('Resolving ' + typename + '...', url)
+                    xml2rfc.log.note('Loaded from cache', cached_path)
+                    return cached_path
+
         xml2rfc.log.note('Resolving ' + typename + '...', url)
         if not netloc in self.sessions:
             self.sessions[netloc] = requests.Session()
@@ -339,21 +362,32 @@ class CachingResolver(lxml.etree.Resolver):
             xml2rfc.log.note("URL retrieval failed with status code %s for '%s'" % (r.status_code, r.url))
             return ''
 
+
+class AnnotatedElement(lxml.etree.ElementBase):
+    pis = None
+
 class XmlRfcParser:
 
+    nsmap = {
+        'xi':   'http://www.w3.org/2001/XInclude',
+    }
+
+
     """ XML parser container with callbacks to construct an RFC tree """
     def __init__(self, source, verbose=False, quiet=False,
                  cache_path=None, templates_path=None, library_dirs=None,
-                 network_locs=[
+                 no_network=False, network_locs=[
+                     'https://xml2rfc.ietf.org/public/rfc/',
+                     'https://xml2rfc.tools.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
-                     'http://xml.resource.org/public/rfc/',
                  ]
                  ):
         self.verbose = verbose
         self.quiet = quiet
         self.source = source
         self.cache_path = cache_path
+        self.no_network = no_network
         self.network_locs = network_locs
 
         # Initialize templates directory
@@ -362,6 +396,9 @@ class XmlRfcParser:
                                            'templates')
         self.default_dtd_path = os.path.join(self.templates_path, 'rfc2629.dtd')
 
+        for prefix, value in self.nsmap.items():
+            lxml.etree.register_namespace(prefix, value)
+
         # If library dirs werent explicitly set, like from the gui, then try:
         #   1. $XML_LIBRARY environment variable as a delimited list
         #   2. Default to /usr/share/xml2rfc
@@ -397,12 +434,20 @@ class XmlRfcParser:
         if not (self.quiet or quiet):
             xml2rfc.log.write('Parsing file', self.source)
 
+        if six.PY2:
+            with open(self.source, "rU") as f:
+                self.text = f.read()
+        else:
+            with open(self.source, "rb", newline=None) as f:
+                self.text = f.read()
+
         # Get an iterating parser object
-        context = lxml.etree.iterparse(self.source,
+        file = six.BytesIO(self.text)
+        context = lxml.etree.iterparse(file,
                                       dtd_validation=False,
                                       load_dtd=True,
                                       attribute_defaults=True,
-                                      no_network=False,
+                                      no_network=self.no_network,
                                       remove_comments=remove_comments,
                                       remove_pis=remove_pis,
                                       remove_blank_text=True,
@@ -416,27 +461,32 @@ class XmlRfcParser:
                                         library_dirs=self.library_dirs,
                                         templates_path=self.templates_path,
                                         source=self.source,
+                                        no_network=self.no_network,
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
-                                    )
+                                     )
         context.resolvers.add(caching_resolver)
 
         # Get hold of the rfc number (if any) in the rfc element, so we can
         # later resolve the "&rfc.number;" entity.
+        self.rfc_number = None
         try:
             for action, element in context:
                 if element.tag == "rfc":
                     self.rfc_number = element.attrib.get("number", None)
+                    break
         except lxml.etree.XMLSyntaxError as e:
             xml2rfc.log.warn("Parsing Error: %s" % e)
-            self.rfc_number = None
+        except ValueError as e:
+            if e.message=="I/O operation on closed file":
+                pass
 
         # now get a regular parser, and parse again, this time resolving entities
         parser = lxml.etree.XMLParser(dtd_validation=False,
                                       load_dtd=True,
                                       attribute_defaults=True,
-                                      no_network=False,
+                                      no_network=self.no_network,
                                       remove_comments=remove_comments,
                                       remove_pis=remove_pis,
                                       remove_blank_text=True,
@@ -448,6 +498,7 @@ class XmlRfcParser:
                                         library_dirs=self.library_dirs,
                                         templates_path=self.templates_path,
                                         source=self.source,
+                                        no_network=self.no_network,
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
@@ -457,17 +508,30 @@ class XmlRfcParser:
         # Add our custom resolver
         parser.resolvers.add(self.cachingResolver)
 
+        # Use our custom element class, which holds the state of PI settings
+        # at this point in the xml tree
+        element_lookup = lxml.etree.ElementDefaultClassLookup(element=AnnotatedElement)
+        parser.set_element_class_lookup(element_lookup)
+
         # Parse the XML file into a tree and create an rfc instance
-        tree = lxml.etree.parse(self.source, parser)
-        xmlrfc = XmlRfc(tree, self.default_dtd_path)
-        
-        # Evaluate processing instructions behind root element
+        file = six.BytesIO(self.text)
+        tree = lxml.etree.parse(file, parser)
+        xmlrfc = XmlRfc(tree, self.default_dtd_path, nsmap=self.nsmap)
+
+        # Evaluate processing instructions before root element
         xmlrfc._eval_pre_pi()
         
-        # Expand 'include' instructions
-        for element in xmlrfc.getroot().iter():
+        # Keep seen elements in a list, to force lxml to not discard (and
+        # recreate) the elements, as this would cause loss of our custom
+        # state, the PI settings at the time the element was parsed
+        # (in element.pis)
+        xmlrfc._elements_cache = []
+        # Process PIs and expand 'include' instructions
+        pis = xmlrfc.pis.copy()
+        for element in xmlrfc.getroot().iterdescendants():
             if element.tag is lxml.etree.PI:
                 pidict = xmlrfc.parse_pi(element)
+                pis = xmlrfc.pis.copy() 
                 if 'include' in pidict and pidict['include']:
                     request = pidict['include']
                     path = self.cachingResolver.getReferenceRequest(request,
@@ -482,8 +546,14 @@ class XmlRfcParser:
                                                       remove_blank_text=True,
                                                       resolve_entities=True,
                                                       strip_cdata=strip_cdata)
+                        parser.set_element_class_lookup(element_lookup)
                         # parser.resolvers.add(self.cachingResolver) --- should this be done?
                         ref_root = lxml.etree.parse(path, parser).getroot()
+                        ref_root.pis = pis
+                        xmlrfc._elements_cache.append(ref_root)
+                        for e in ref_root.iterdescendants():
+                            e.pis = pis
+                            xmlrfc._elements_cache.append(e)
                         parent = element.getparent()
                         parent.replace(element, ref_root)
                     except (lxml.etree.XMLSyntaxError, IOError) as e:
@@ -494,7 +564,10 @@ class XmlRfcParser:
                         else:
                             xml2rfc.log.warn('Unable to load the include file at',
                                               path)
-
+            else:
+                if isinstance(element, AnnotatedElement):
+                    element.pis = pis
+                    xmlrfc._elements_cache.append(element)                    
         # Finally, do any extra formatting on the RFC before returning
         xmlrfc._format_whitespace()
 
@@ -509,13 +582,16 @@ class XmlRfc:
         Accessing the rfc tree is done by getting the root node from getroot()
     """
 
-    def __init__(self, tree, default_dtd_path):
+    pis = {}
+
+    def __init__(self, tree, default_dtd_path, nsmap={}):
         self.default_dtd_path = default_dtd_path
         self.tree = tree
+        self.nsmap = nsmap
         # Pi default values
         self.pis = {
             "artworkdelimiter":	None,
-            "artworklines":	0 ,
+            "artworklines":	"0",
             "authorship":	"yes",
             "autobreaks":	"yes",
             "background":	"" ,
@@ -523,37 +599,40 @@ class XmlRfc:
             "comments":		"no" ,
             "docmapping":	"no",
             "editing":		"no",
-            "emoticonic":	"no",
+            #"emoticonic":	"no",
             #"footer":		Unset
             "figurecount":      "no",
             #"header":		Unset
             "inline":		"no",
-            "iprnotified":	"no",
+            #"iprnotified":	"no",
             "linkmailto":	"yes",
             #"linefile":	Unset
-            #"needLines":       Unset
+            "needLines":        None,
             "multiple-initials":"no",
-            "notedraftinprogress": "yes",
+            #"notedraftinprogress": "yes",
+            "orphanlimit":      "2",
             "private":		"",
             "refparent":	"References",
             "rfcedstyle":	"no",
-            "rfcprocack":	"no",
-            "sectionorphan":    4,
-            "slides":		"no",
-            "sortrefs":		"no",
-            "strict":		"no",
+            #"rfcprocack":	"no",
+            "sectionorphan":    "4",
+            #"slides":		"no",
+            "sortrefs":		"yes",
+            #"strict":		"no",
             "symrefs":		"yes",
             "tablecount":       "no",
             "text-list-symbols": "o*+-",
             "toc":		"no",
             "tocappendix":	"yes",
-            "tocdepth":		3,
+            "tocdepth":		"3",
             "tocindent":	"yes",
             "tocnarrow":	"yes",
-            "tocompact":	"yes",
+            #"tocompact":	"yes",
+            "tocpagebreak":     "no",
             "topblock":		"yes",
             #"typeout":		Unset
-            "useobject":	"no" ,
+            #"useobject":	"no" ,
+            "widowlimit":       "2",
         }
         # Special cases:
         self.pis["compact"] = self.pis["rfcedstyle"]
@@ -565,7 +644,7 @@ class XmlRfc:
 
     def getpis(self):
         """ Returns a list of the XML processing instructions """
-        return self.pis
+        return self.pis.copy()
     
     def validate(self, dtd_path=None):
         """ Validate the document with its default dtd, or an optional one 
@@ -605,24 +684,9 @@ class XmlRfc:
             else:
                 # The document was not valid
                 return False, dtd.error_log
-    
+
     def parse_pi(self, pi):
-        """ Add a processing instruction to the current state 
-            
-            Will also return the dictionary containing the added instructions
-            for use in things like ?include instructions
-        """
-        if pi.text:
-            # Split text in the format 'key="val"'
-            chunks = re.split(r'=[\'"]([^\'"]*)[\'"]', pi.text)
-            # Create pairs from this flat list, discard last element if odd
-            tmp_dict = dict(zip(chunks[::2], chunks[1::2]))
-            for key, val in tmp_dict.items():
-                # Update main PI state
-                self.pis[key] = val
-            # Return the new values added
-            return tmp_dict
-        return {}
+        return xml2rfc.utils.parse_pi(pi, self.pis)
 
     def _eval_pre_pi(self):
         """ Evaluate pre-document processing instructions
