@@ -6,7 +6,8 @@ from xml2rfc.writers.paginated_txt import PaginatedTextRfcWriter
 from xml2rfc.writers.html import HtmlRfcWriter
 from xml2rfc.writers.nroff import NroffRfcWriter
 from xml2rfc.writers.expanded_xml import ExpandedXmlWriter
+from xml2rfc.writers.v2v3 import V2v3XmlWriter
 
 __all__ = ['BaseRfcWriter', 'RawTextRfcWriter', 'PaginatedTextRfcWriter',
            'HtmlRfcWriter', 'NroffRfcWriter', 'ExpandedXmlWriter',
-           'RfcWriterError',]
+           'RfcWriterError', 'V2v3XmlWriter', ]
