@@ -9,6 +9,7 @@ import lxml
 import re
 import xml2rfc.log
 import xml2rfc.utils
+from optparse import Values
 
 try:
     import debug
@@ -16,6 +17,8 @@ try:
 except ImportError:
     pass
 
+default_options = Values(defaults=dict(quiet=False, verbose=False, utf8=False, debug=False))
+
 class _RfcItem:
     """ A unique ID object for an anchored RFC element.
     
@@ -87,6 +90,9 @@ class BaseRfcWriter:
     # Boilerplate text
     boilerplate = {}
 
+    # The RFC Editor's switchover date for https URLs
+    boilerplate_https_date = datetime.datetime(year=2017, month=8, day=21)
+
     # Document stream names
     boilerplate['document_stream'] = {}
     boilerplate['document_stream']['IETF'] = \
@@ -125,6 +131,9 @@ class BaseRfcWriter:
         'info': {},
         'historic': {},
     }
+    # 'Status of this Memo' boilerplate for RFCs
+    boilerplate['status_5741'] = {
+    }
 
     # Paragraph 1
     boilerplate['status']['std']['p1'] = \
@@ -190,13 +199,26 @@ class BaseRfcWriter:
     # Paragraph 2 last sentence
     boilerplate['status']['p2end_ietf_std'] = \
         'Further information on Internet Standards is available ' \
-        'in Section 2 of RFC 5741.'
+        'in Section 2 of RFC 7841.'
     boilerplate['status']['p2end_ietf_bcp'] = \
-        'Further information on BCPs is available in Section 2 of RFC 5741.'
+        'Further information on BCPs is available in Section 2 of RFC 7841.'
     boilerplate['status']['p2end_ietf_other'] = \
         'Not all documents approved by the IESG are a candidate for any ' \
-        'level of Internet Standard; see Section 2 of RFC 5741.'
+        'level of Internet Standard; see Section 2 of RFC 7841.'
     boilerplate['status']['p2end_other'] = \
+        'Documents approved for publication by the %s are not a ' \
+        'candidate for any level of Internet Standard; see Section 2 of RFC ' \
+        '7841.'
+
+    boilerplate['status_5741']['p2end_ietf_std'] = \
+        'Further information on Internet Standards is available ' \
+        'in Section 2 of RFC 5741.'
+    boilerplate['status_5741']['p2end_ietf_bcp'] = \
+        'Further information on BCPs is available in Section 2 of RFC 5741.'
+    boilerplate['status_5741']['p2end_ietf_other'] = \
+        'Not all documents approved by the IESG are a candidate for any ' \
+        'level of Internet Standard; see Section 2 of RFC 5741.'
+    boilerplate['status_5741']['p2end_other'] = \
         'Documents approved for publication by the %s are not a ' \
         'candidate for any level of Internet Standard; see Section 2 of RFC ' \
         '5741.'
@@ -206,6 +228,10 @@ class BaseRfcWriter:
         'Information about the current status of this document, any errata, '
         'and how to provide feedback on it may be obtained at '
         'http://www.rfc-editor.org/info/rfc%s.')
+    boilerplate['status']['p3_s'] = (
+        'Information about the current status of this document, any errata, '
+        'and how to provide feedback on it may be obtained at '
+        'https://www.rfc-editor.org/info/rfc%s.')
 
     # 'Status of this Memo' boilerplate for drafts
     boilerplate['status']['draft'] = [
@@ -218,6 +244,17 @@ class BaseRfcWriter:
        'and may be updated, replaced, or obsoleted by other documents at any '
        'time.  It is inappropriate to use Internet-Drafts as reference '
        'material or to cite them other than as "work in progress."']
+    boilerplate['status']['draft_s'] = [
+       'Internet-Drafts are working documents of the Internet Engineering '
+       'Task Force (IETF).  Note that other groups may also distribute '
+       'working documents as Internet-Drafts.  The list of current Internet-'
+       'Drafts is at https://datatracker.ietf.org/drafts/current/.',
+       #
+       'Internet-Drafts are draft documents valid for a maximum of six months '
+       'and may be updated, replaced, or obsoleted by other documents at any '
+       'time.  It is inappropriate to use Internet-Drafts as reference '
+       'material or to cite them other than as "work in progress."']
+
     boilerplate['draft_expire'] = \
        'This Internet-Draft will expire on %s.'
 
@@ -240,6 +277,13 @@ class BaseRfcWriter:
         'publication of this document.  Please review these documents '
         'carefully, as they describe your rights and restrictions with respect '
         'to this document.')
+    boilerplate['base_copyright_body_s'] = (
+        'This document is subject to BCP 78 and the IETF Trust\'s Legal '
+        'Provisions Relating to IETF Documents '
+        '(https://trustee.ietf.org/license-info) in effect on the date of '
+        'publication of this document.  Please review these documents '
+        'carefully, as they describe your rights and restrictions with respect '
+        'to this document.')
 
     # IPR values which append things to copyright
     boilerplate['ipr_200902_copyright_ietfbody'] = \
@@ -296,9 +340,10 @@ class BaseRfcWriter:
 
     # -------------------------------------------------------------------------
 
-    def __init__(self, xmlrfc, quiet=False, verbose=False, date=datetime.date.today()):
-        self.quiet = quiet
-        self.verbose = verbose
+    def __init__(self, xmlrfc, quiet=None, options=default_options, date=datetime.datetime.now()):
+        if not quiet is None:
+            options.quiet = quiet
+        self.options = options
         self.date = date
         self.expire_string = ''
         self.ascii = False
@@ -441,17 +486,35 @@ class BaseRfcWriter:
         """author is an rfc2629 author element.  Return the author initials,
         fixed up according to current flavour and policy."""
         initials = author.attrib.get('initials', '')
+        multiple = author.pis["multiple-initials"] == "yes"
         initials_list = re.split("[. ]+", initials)
-        if self.pis["multiple-initials"] == "no":
-            if not initials_list[0] == "":
+        try:
+            initials_list.remove('')
+        except:
+            pass
+        if len(initials_list) > 0:
+            if multiple:
+                # preserve spacing, but make sure all parts have a trailing
+                # period
+                initials = initials.strip()
+                initials += '.' if not initials.endswith('.') else ''
+                initials = re.sub('([^.]) ', '\g<1>. ', initials)
+            else:
                 initials = initials_list[0] + "."
-        else:
-            initials = ". ".join(initials_list) + "."
         return initials
 
     def parse_pi(self, pi):
         return xml2rfc.utils.parse_pi(pi, self.pis)
 
+    def get_numeric_pi(self, key, default):
+        num = self.pis.get(key, None)
+        if num is None:
+            return default
+        if not num.isdigit():
+            xml2rfc.log.warn('Expected a numeric value for the %s PI, found "%s"' % (key, num))
+            return default
+        return int(num)
+
     def _getTocIndex(self):
         return [item for item in self._index if item.toc]
         
@@ -468,56 +531,69 @@ class BaseRfcWriter:
             raise RfcWriterError('No boilerplate text available for '
             'ipr: \'%s\'.  Acceptable values are: ' % ipr + \
             ', '.join(self.supported_ipr))
+        
+    def is_numbered(self, node):
+        attr = node.attrib.get('numbered', 'true')
+        return attr in ['yes', 'true', ]
     
     def _format_date(self):
         """ Fix the date data """
         today = self.date
         date = self.r.find('front/date')
-        if date is not None:
-            year = date.attrib.get('year', '')
-            month = date.attrib.get('month', '')
-            day = date.attrib.get('day', '')
-            if not year or (year == str(today.year) and\
-               (not month or month.lower() == today.strftime("%B").lower() or \
-                month.lower() == today.strftime("%b").lower())):
-                # Set everything to today
-                date.attrib['year'] = today.strftime('%Y')
-                date.attrib['month'] = today.strftime('%B')
-                if self.draft and not day:
-                    date.attrib['day'] = today.strftime('%d')
-                    if date.attrib['day'][0] == '0':
-                        date.attrib['day'] = today.strftime('%d').replace('0', '')
-            elif year != str(today.year) and not month:
-                xml2rfc.log.error("Incomplete and out-of date <date/> element: %s" % lxml.etree.tostring(date))
+        assert date is not None, "Bug in schema validation: no date element in document"
+
+        year = date.attrib.get('year', '')
+        month = date.attrib.get('month', '')
+        day = date.attrib.get('day', '')
+        if not year or (year == str(today.year) and\
+           (not month or month.lower() == today.strftime("%B").lower() or \
+            month.lower() == today.strftime("%b").lower())):
+            # Set everything to today
+            date.attrib['year'] = today.strftime('%Y')
+            date.attrib['month'] = today.strftime('%B')
+            if self.draft and not day:
+                date.attrib['day'] = today.strftime('%d')
+                if date.attrib['day'][0] == '0':
+                    date.attrib['day'] = today.strftime('%d').replace('0', '')
+        elif year != str(today.year) and not month:
+            xml2rfc.log.error("Incomplete and out-of date <date/> element: %s" % lxml.etree.tostring(date))
+
         try:
+            # Full month name
             datetime.datetime.strptime(date.attrib.get('year')+date.attrib.get('month'), '%Y%B')
         except ValueError:
             try:
-                datetime.datetime.strptime(date.attrib.get('year')+date.attrib.get('month'), '%Y%b')
+                # Abbreviated month name
+                d = datetime.datetime.strptime(date.attrib.get('year')+date.attrib.get('month'), '%Y%b')
+                date.attrib['month'] = d.strftime('%B')
             except ValueError:
                 xml2rfc.log.warn("Year and/or month are incorrect values in <date/> element: %s" % lxml.etree.tostring(date))
         except TypeError:
             pass
 
+        self.date = datetime.datetime.strptime(date.attrib['year']+date.attrib['month']+date.attrib.get('day','1'), '%Y%B%d')
+
         # Setup the expiration string for drafts as published date + six months
+        # FIXME: The following code undoes some of the code above.  Determine
+        # which should actually be in effect, and get rid of the other.
         if self.draft:
             date = self.r.find('front/date')
-            if date is not None:
-                month = date.attrib.get('month', '')
-                year = date.attrib.get('year', '')
-                day = date.attrib.get('day', '1')  # Default to first of month
+
+            month = date.attrib.get('month', '')
+            year = date.attrib.get('year', '')
+            day = date.attrib.get('day', '1')  # Default to first of month
+            try:
+                start_date = datetime.datetime.strptime(year + month + day, \
+                                                        '%Y%B%d')
+            except ValueError:
                 try:
                     start_date = datetime.datetime.strptime(year + month + day, \
-                                                            '%Y%B%d')
+                                                            '%Y%b%d')
                 except ValueError:
-                    try:
-                        start_date = datetime.datetime.strptime(year + month + day, \
-                                                                '%Y%b%d')
-                    except ValueError:
-                        start_date = today
+                    start_date = today
 
-                expire_date = start_date + datetime.timedelta(185)
-                self.expire_string = expire_date.strftime('%B %d, %Y').replace(' 0', ' ')
+            expire_date = start_date + datetime.timedelta(185)
+            self.expire_string = expire_date.strftime('%B %d, %Y').replace(' 0', ' ')
 
 
     def _format_counter(self, text, count, list_length=1):
@@ -572,15 +648,22 @@ class BaseRfcWriter:
                 initials = self.get_initials(author)
                 if i == len(authors) - 1 and len(authors) > 1:
                     # Last author is rendered in reverse
-                    buf.append(initials + ' ' + \
+                    if len(initials) > 0:
+                        buf.append(initials + ' ' + \
                                      surname)
-                else:
+                    else:
+                        buf.append(surname)
+                elif len(initials) > 0:
                     buf.append(surname + ', ' + initials)
+                else:
+                    buf.append(surname)
                 if author.attrib.get('role', '') == 'editor':
                     buf.append(', Ed.')
             elif organization is not None and organization.text:
                 # Use organization instead of name
                 buf.append(organization.text.strip())
+            else:
+                continue
             if len(authors) == 2 and i == 0:
                 buf.append(' ')
             elif i < len(authors) - 1:
@@ -657,20 +740,20 @@ class BaseRfcWriter:
         """ Returns a list of lines for the top right header """
         lines = []
         # Render author?
-        authorship = self.pis['authorship']
-        if authorship == 'yes':
-            # Keep track of previous organization and remove if redundant.
-            last_org = None
-            last_pos = None
-            for author in self.r.findall('front/author'):
-                role = author.attrib.get('role', '')
-                if role == 'editor':
-                    role = ', Ed.'
-                initials = self.get_initials(author)
-                lines.append(initials + ' ' + author.attrib.\
-                             get('surname', '') + role)
-                organization = author.find('organization')
-                org_name = ''
+        stream = self.r.attrib.get('submissionType', self.defaults['submissionType'])
+        # Keep track of previous organization and remove if redundant.
+        last_org = None
+        last_pos = None
+        for author in self.r.findall('front/author'):
+            role = author.attrib.get('role', '')
+            if role == 'editor':
+                role = ', Ed.'
+            initials = self.get_initials(author)
+            lines.append(initials + ' ' + author.attrib.\
+                         get('surname', '') + role)
+            organization = author.find('organization')
+            org_name = ''
+            if stream != 'IAB':
                 if organization is not None:
                     abbrev = organization.attrib.get("abbrev", None)
                     if  abbrev != None and abbrev.strip() != '':
@@ -684,12 +767,12 @@ class BaseRfcWriter:
                         # Remove redundant organization
                         del lines[last_pos]
                     lines.append(org_name)
-                last_org = org_name
-                last_pos = len(lines)-1
-            # remove blank lines between authors and date
-            if lines[last_pos] == '':
-                del lines[last_pos]
-                last_pos = len(lines)-1
+            last_org = org_name
+            last_pos = len(lines)-1
+        # remove blank lines between authors and date
+        if lines[last_pos] == '':
+            del lines[last_pos]
+            last_pos = len(lines)-1
 
         date = self.r.find('front/date')
         if date is not None:
@@ -735,10 +818,10 @@ class BaseRfcWriter:
                 self._add_iref_to_index(element)
 
         # Write figure with optional delimiter
-        delimiter = self.pis['artworkdelimiter']
+        delimiter = figure.pis['artworkdelimiter']
         artwork = figure.find('artwork')
         artwork_align = artwork.attrib.get('align', figure_align)
-        blanklines = int(self.pis['artworklines'])
+        blanklines = int(figure.pis['artworklines'])
         self.write_raw(artwork.text, align=artwork_align,
                        blanklines=blanklines, delimiter=delimiter,
                        source_line=figure.sourceline)
@@ -807,8 +890,17 @@ class BaseRfcWriter:
         pass
         
 
+    def numToAlpha(self, n):
+        str = ""
+        az = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+        while n > 0:
+            r = (n-1) % 26
+            n = int((n-1)/26)
+            str = az[r] + str
+        return str
+
     def write_section_rec(self, section, count_str="1.", appendix=False,
-                           level=0, numbered=True):
+                           level=0, numbered=True, emit_numbered="all", s_count=1):
         """ Recursively writes <section> elements 
         
             We use the self.indexmode flag to determine whether or not we
@@ -821,12 +913,12 @@ class BaseRfcWriter:
             include_toc = section.attrib.get('toc', self.defaults['section_toc']) != 'exclude' \
                           and (not appendix or self.pis['tocappendix'] == 'yes')
             if level == 1:
-                numbered = numbered and section.attrib.get('numbered') != "no"
+                numbered = numbered and self.is_numbered(section)
             else:
-                numbered = section.attrib.get('numbered') != "no"
+                numbered = self.is_numbered(section)
             if level > 1 and not numbered:
                 if not self.indexmode:
-                    xml2rfc.log.warn('Unnumbered subsections are not permitted: found section "%s" with attribute numbered="no"' % (title, ))
+                    xml2rfc.log.warn('Unnumbered subsections are not permitted: found section "%s" with attribute numbered="no" or "false"' % (title, ))
                 numbered = True
             if self.indexmode:
                 # Add section to the index
@@ -870,8 +962,6 @@ class BaseRfcWriter:
             elif element.tag == 'iref':
                 self._add_iref_to_index(element)
 
-        s_count = 1  # Section counter
-        
         # Append a dot to separate sub counters
         if count_str:
             count_str += '.'
@@ -880,18 +970,21 @@ class BaseRfcWriter:
         for child_sec in section.findall('section'):
             if level == 0:
                 if numbered:
-                    numbered = child_sec.attrib.get('numbered') != "no"
-                elif child_sec.attrib.get('numbered') != "no":
+                    numbered = self.is_numbered(child_sec)
+                elif self.is_numbered(child_sec):
                     title = child_sec.attrib.get('title')
                     if not self.indexmode:
                         xml2rfc.log.warn('Numbered sections are not permitted after unnumbered sections: found section "%s" without attribute numbered="no"' % (title,))
                     numbered = False
+                if emit_numbered=="only" and not numbered:
+                    continue
+                if emit_numbered=="no" and numbered:
+                    continue
             if appendix == True and not count_str:
                 if s_count == 1 and self.pis["rfcedstyle"] == "yes":
                    self.needLines(-1)
                 # Use an alphabetic counter for first-level appendix
-                uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-                self.write_section_rec(child_sec, uppercase[s_count - 1],
+                self.write_section_rec(child_sec, self.numToAlpha(s_count),
                                         level=level + 1, appendix=True, numbered=numbered)
             else:
                 # Use a numeric counter
@@ -904,6 +997,8 @@ class BaseRfcWriter:
         if count_str == '' and appendix == False:
             self.refs_start = s_count
 
+        return s_count
+
     def write_status_section(self):
         """ Writes the 'Status of this Memo' section """
 
@@ -952,21 +1047,27 @@ class BaseRfcWriter:
                 p2.append(self.boilerplate['status'].get(stream, ''))
 
             # Last sentence of p2
-            if stream == 'IETF' and category == 'std':
-                p2.append(self.boilerplate['status']['p2end_ietf_std'])
-            elif stream == 'IETF' and category == 'bcp':
-                p2.append(self.boilerplate['status']['p2end_ietf_bcp'])
-            elif stream == 'IETF':
-                p2.append(self.boilerplate['status']['p2end_ietf_other'])
+            if self.date < datetime.datetime(year=2016, month=6, day=1):
+                status="status_5741"
             else:
-                p2.append(self.boilerplate['status']['p2end_other'] \
+                status="status"         # Current boilerplate
+            if stream == 'IETF' and category == 'std':
+                p2.append(self.boilerplate[status]['p2end_ietf_std'])
+            elif stream == 'IETF' and category == 'bcp':
+                p2.append(self.boilerplate[status]['p2end_ietf_bcp'])
+            elif stream == 'IETF':
+                p2.append(self.boilerplate[status]['p2end_ietf_other'])
+            else:
+                p2.append(self.boilerplate[status]['p2end_other'] \
                           % self.approvers.get(stream, ''))
 
+
             # Write second paragraph
             self.write_paragraph('  '.join(p2))
             
             # Write third paragraph
-            self.write_paragraph(self.boilerplate['status']['p3'] % self.rfcnumber)
+            key = 'p3' if self.date < self.boilerplate_https_date else 'p3_s'
+            self.write_paragraph(self.boilerplate['status'][key] % self.rfcnumber)
 
         else:  # Draft
             # Start by checking for an ipr header
@@ -977,7 +1078,8 @@ class BaseRfcWriter:
                 self.write_paragraph(self.boilerplate['ipr_200811_status'])
 
             # Write the standard draft status
-            for par in self.boilerplate['status']['draft']:
+            key = 'draft' if self.date < self.boilerplate_https_date else 'draft_s'
+            for par in self.boilerplate['status'][key]:
                 self.write_paragraph(par)
 
             # Write expiration string, if it was generated
@@ -998,7 +1100,8 @@ class BaseRfcWriter:
 
         # Write next paragraph which may be modified by ipr
         ipr = self.r.attrib.get('ipr', self.defaults['ipr'])
-        body = self.boilerplate['base_copyright_body']
+        key = 'base_copyright_body' if self.date < self.boilerplate_https_date else 'base_copyright_body_s'
+        body = self.boilerplate[key]
         if '200902' in ipr and self.r.attrib.get('submissionType', 
                                    self.defaults['submissionType']) == 'IETF':
             body += '  ' + self.boilerplate['ipr_200902_copyright_ietfbody']
@@ -1029,6 +1132,12 @@ class BaseRfcWriter:
             for t in abstract.findall('t'):
                 self.write_t_rec(t)
 
+        # Optional notes
+        for note in self.r.findall('front/note'):
+            self.write_heading(note.attrib.get('title', 'Note'))
+            for t in note.findall('t'):
+                self.write_t_rec(t)
+
         # Middle sections
         middle = self.r.find('middle')
         if middle is not None:
@@ -1067,7 +1176,7 @@ class BaseRfcWriter:
         # Appendix sections
         back = self.r.find('back')
         if back is not None:
-            self.write_section_rec(back, None, appendix=True)
+            s_count = self.write_section_rec(back, None, appendix=True, emit_numbered="only")
 
         # Index section, disable if there are no irefs
         if len(self._iref_index) > 0:
@@ -1077,16 +1186,21 @@ class BaseRfcWriter:
             item = _RfcItem(title, autoAnchor, title=title)
             self._index.append(item)
 
+        if back is not None:
+            self.write_section_rec(back, None, appendix=True, emit_numbered="no", s_count=s_count)
+
         # Authors addresses section
-        authors = self.r.findall('front/author')
-        autoAnchor = 'rfc.authors'
-        if len(authors) > 1:
-            title = "Authors' Addresses"
-        else:
-            title = "Author's Address"
-        # Add explicitly to index
-        item = _RfcItem(title, autoAnchor, title=title)
-        self._index.append(item)
+
+        if self.pis['authorship'] == 'yes':
+            authors = self.r.findall('front/author')
+            autoAnchor = 'rfc.authors'
+            if len(authors) > 1:
+                title = "Authors' Addresses"
+            else:
+                title = "Author's Address"
+            # Add explicitly to index
+            item = _RfcItem(title, autoAnchor, title=title)
+            self._index.append(item)
 
     def _build_document(self):
         self.indexmode = False
@@ -1186,24 +1300,28 @@ class BaseRfcWriter:
         # Appendix sections
         back = self.r.find('back')
         if back is not None:
-            self.write_section_rec(back, None, appendix=True)
-
-        self.write_crefs()
+            s_count = self.write_section_rec(back, None, appendix=True, emit_numbered="only")
 
         # Index section, disable if there are no irefs
         if len(self._iref_index) > 0:
             self.insert_iref_index()
 
+        if back is not None:
+            self.write_section_rec(back, None, appendix=True, emit_numbered="no", s_count=s_count)
+
+        self.write_crefs()
+
         # Authors addresses section
-        authors = self.r.findall('front/author')
-        autoAnchor = 'rfc.authors'
-        if len(authors) > 1:
-            title = "Authors' Addresses"
-        else:
-            title = "Author's Address"
-        self.write_heading(title, autoAnchor=autoAnchor)
-        for author in authors:
-            self.write_address_card(author)
+        if self.pis['authorship'] == 'yes':
+            authors = self.r.findall('front/author')
+            autoAnchor = 'rfc.authors'
+            if len(authors) > 1:
+                title = "Authors' Addresses"
+            else:
+                title = "Author's Address"
+            self.write_heading(title, autoAnchor=autoAnchor)
+            for author in authors:
+                self.write_address_card(author)
 
         self.check_for_unused_references()
         
@@ -1213,6 +1331,9 @@ class BaseRfcWriter:
         if self.ascii:
             xml2rfc.utils.safeReplaceUnicode(self.r)
 
+        # Protect any words with slashes that are on the preserve list
+        xml2rfc.utils.safeTagSlashedWords(self.r)
+
         # Do any pre processing necessary, such as inserting metadata
         self.pre_indexing()
         # Make two passes over the document, the first pass we run in
@@ -1242,7 +1363,7 @@ class BaseRfcWriter:
         else:
             self.write_to_file(tmpfile)
 
-        if not self.quiet and filename:
+        if not self.options.quiet and filename:
             xml2rfc.log.write('Created file', filename)
 
     def write_erefs(self, refs_counter, refs_subsection):
