@@ -10,6 +10,7 @@ import string
 import sys
 import datetime
 import cgi
+
 try:
     import debug
     assert debug
@@ -18,7 +19,7 @@ except ImportError:
 
 # Local libs
 import xml2rfc
-from xml2rfc.writers.base import BaseRfcWriter
+from xml2rfc.writers.base import BaseRfcWriter, default_options
 
 
 class HtmlRfcWriter(BaseRfcWriter):
@@ -28,11 +29,14 @@ class HtmlRfcWriter(BaseRfcWriter):
     """
     # HTML Specific Defaults that are not provided in templates or XML
     html_defaults = { 
-        'references_url': 'http://tools.ietf.org/html/',
+        'references_url': 'https://tools.ietf.org/html/',
     }
 
-    def __init__(self, xmlrfc, quiet=False, verbose=False, date=datetime.date.today(), templates_dir=None):
-        BaseRfcWriter.__init__(self, xmlrfc, quiet=quiet, date=date, verbose=verbose)
+    def __init__(self, xmlrfc, quiet=None, options=default_options,
+                               date=datetime.date.today(), templates_dir=None):
+        if not quiet is None:
+            options.quiet = quiet
+        BaseRfcWriter.__init__(self, xmlrfc, options=options)
         self.list_counters = {}
         self.iref_index = []
         self.nbws_cond = ' '    # Use space as they are in an anchor element anyway
@@ -184,10 +188,10 @@ class HtmlRfcWriter(BaseRfcWriter):
                 link.attrib['rel'] = 'copyright' in item.autoAnchor and \
                                      'Copyright' or 'Chapter'
                 if item.title and item.counter:
-                    link.attrib['title'] = item.counter + ' ' + item.title
+                    link.attrib['title'] = item.counter + ' ' + item.title.strip()
                 self.buffers['toc_head_links'].append(self._serialize(link))
                 # Create actual toc list item
-                a = E.A(item.title, href='#' + item.autoAnchor)
+                a = E.A(item.title.strip(), href='#' + item.autoAnchor)
                 counter_text = item.counter and item.counter + '.   ' or ''
                 # Prepend appendix at first level
                 if item.level == 1 and item.appendix:
@@ -200,9 +204,9 @@ class HtmlRfcWriter(BaseRfcWriter):
 
     def _serialize(self, element):
         if sys.version > '3':
-            return lxml.html.tostring(element, pretty_print=True, method='xml', encoding='ascii').decode()
+            return lxml.html.tostring(element, pretty_print=True, method='html', encoding='utf-8').decode('utf-8')
         else:
-            return lxml.html.tostring(element, pretty_print=True, method='xml')
+            return lxml.html.tostring(element, pretty_print=True, method='html')
     
     def _flush_temp_div(self):
         lines = []
@@ -215,48 +219,51 @@ class HtmlRfcWriter(BaseRfcWriter):
         """ Return a list of HTML elements that represent the reference """   
         if element.tag == 'xref':
             target = element.attrib.get('target', '')
+            format = element.attrib.get('format', self.defaults['xref_format'])
             item = self._getItemByAnchor(target)
             if not self.indexmode:
                 if not item:
                     xml2rfc.log.warn("Can't resolve xref target %s" % target)
                 else:
                     item.used = True
-            if element.text:
-                cite = E.CITE('[' + target + ']', title='NONE')
-                a = E.A(element.text, href='#' + target)
-                a.tail = ' '
-                if element.tail:
-                    cite.tail = element.tail
-                return [a, cite]
+            # Create xref from index lookup
+            if not item:
+                text = '[' + target + ']'
+            elif format == 'none':
+                text = ''
+            elif format == 'counter':
+                text = item.counter
+            elif format == 'title':
+                text = item.title.strip()
             else:
-                # Create xref from index lookup
-                format = element.attrib.get('format', 
-                                            self.defaults['xref_format'])
-                a = E.A(href='#' + target)
-                if not item:
-                    text = '[' + target + ']'
-                elif format == 'none':
-                    text = ''
-                elif format == 'counter':
-                    text = item.counter
-                elif format == 'title':
-                    text = item.title
-                else:
-                    # Default
-                    text = item.autoName
-                a.text = text
-                if element.tail:
-                    a.tail = element.tail
-                return [a]
+                # Default
+                text = item.autoName
+
+            # following the V3 HTML -
+            #  If you specify text, that is what you get.
+            if element.text:
+                text = element.text.rstrip()
+             
+            a = E.A(href='#' + target)
+            a.attrib["class"] = "xref"
+            a.text = text
+            if element.tail:
+                a.tail = element.tail
+            
+            return [a]
+
         elif element.tag == 'eref':
             target = element.attrib.get('target', '')
-            text = element.text or target
-            if text:
-                a = E.A(text, href=target)
+            if element.text:
+                a = E.A(element.text, href=target)
                 a.tail = element.tail
-#                cite = E.CITE('[' + target + ']', title='NONE')
-#                current.append(cite)
                 return [a]
+            else:
+                sp1 = E.SPAN('<')
+                a = E.A(target, href=target)
+                sp2 = E.SPAN('>')
+                sp2.tail = element.tail
+                return [sp1, a, sp2]
         elif element.tag == 'cref':
             self.cref_counter += 1
             anchor = element.attrib.get('anchor', None)
@@ -346,7 +353,7 @@ class HtmlRfcWriter(BaseRfcWriter):
             h.append(a_bullet)
             if anchor:
                 # Use an anchor link for heading
-                a_text = E.A(text, href='#' + anchor, )
+                a_text = E.A(text.strip(), href='#' + anchor, )
                 a_text.attrib["id"] = anchor
                 h.append(a_text)
             else:
@@ -535,37 +542,42 @@ class HtmlRfcWriter(BaseRfcWriter):
             bullet_td.attrib['class'] = 'reference'
             ref_td = E.TD()
             ref_td.attrib['class'] = 'top'
-            last = ref_td
+            last = None
             authors = reference.findall('front/author')
             for j, author in enumerate(authors):
                 organization = author.find('organization')
                 email = author.find('address/email')
                 surname = author.attrib.get('surname')
                 initials = self.get_initials(author)
+                a = None
                 if j == len(authors) - 1 and len(authors) > 1:
                     last.tail = ' and '
                 if surname is not None:
                     if j == len(authors) - 1 and len(authors) > 1:
                         # Last author, render in reverse
-                        name_string = initials + ' ' + surname 
+                        if len(initials) > 0:
+                            name_string = initials + ' ' + surname
+                        else:
+                            name_string = surname
                     else:
-                        name_string = surname + ', ' + initials
+                        if len(initials) > 0:
+                            name_string = surname + ', ' + initials
+                        else:
+                            name_string = surname
                     a = E.A(name_string)
                     if email is not None and email.text:
                         if self.pis['linkmailto'] == 'yes':
                             a.attrib['href'] = 'mailto:' + email.text
                     if organization is not None and organization.text:
                         a.attrib['title'] = organization.text.strip()
-                    ref_td.append(a)
-                    last = a
                 elif organization is not None and organization.text:
                     # Use organization instead of name
                     a = E.A(organization.text.strip())
-                    ref_td.append(a)
-                    last = a
+                else:
+                    continue
+                ref_td.append(a)
+                last = a
                 a.tail = ', '
-            if len(authors):
-                last.tail = ', "'
             title = reference.find('front/title')
             if title is not None and title.text:
                 title_string = title.text.strip()
@@ -573,9 +585,20 @@ class HtmlRfcWriter(BaseRfcWriter):
                 xml2rfc.log.warn('No title specified in reference', \
                                  reference.attrib.get('anchor', ''))
                 title_string = ''
-            title_a = E.A(title_string)
-            title_a.tail = '", '
-            ref_td.append(title_a)
+            if title_string:
+                if reference.attrib.get("quote-title", "true") == "true": # attribute default value: yes
+                    if last is not None:
+                        last.tail = ', "' 
+                    else:
+                        ref_td.text = '"'
+                    title_a = E.A(title_string)
+                    title_a.tail = '"'
+                else:
+                    if last is not None:
+                        last.tail = ', '
+                    title_a = E.A(title_string)
+                    title_a.tail = ''
+                ref_td.append(title_a)
             for seriesInfo in reference.findall('seriesInfo'):
                 # Create title's link to document from seriesInfo
                 if seriesInfo.attrib.get('name', '') == 'RFC':
@@ -586,8 +609,8 @@ class HtmlRfcWriter(BaseRfcWriter):
                     title_a.attrib['href'] = \
                         self.html_defaults['references_url'] + \
                         seriesInfo.attrib.get('value', '')
-                title_a.tail += seriesInfo.attrib.get('name', '') + ' ' + \
-                             seriesInfo.attrib.get('value', '') + ', '
+                title_a.tail += ', '+seriesInfo.attrib.get('name', '') + ' ' + \
+                             seriesInfo.attrib.get('value', '')
             if not title_a.attrib.has_key("href"):
                 href = reference.attrib.get("target", None)
                 if href:
@@ -595,10 +618,12 @@ class HtmlRfcWriter(BaseRfcWriter):
             date = reference.find('front/date')
             if date is not None:
                 month = date.attrib.get('month', '')
-                if month:
-                    month += ' '
                 year = date.attrib.get('year', '')
-                title_a.tail += month + year + '.'
+                if month or year:
+                    title_a.tail += ', '
+                    if month:
+                        month += ' '
+                    title_a.tail += month + year + '.'
             tr.append(bullet_td)
             tr.append(ref_td)
             refdict[bullet] = tr
@@ -606,9 +631,12 @@ class HtmlRfcWriter(BaseRfcWriter):
             # Render annotation as a separate paragraph
             annotation = reference.find('annotation')
             if annotation is not None and annotation.text:
-                ref_td.append(E.P(annotation.text))
+                p = E.P()
+                self.write_t_rec(annotation, parent=p)
+                ref_td.append(p)
+                    
         if self.pis['sortrefs'] == 'yes' and self.pis['symrefs'] == 'yes':
-            refkeys = sorted(refkeys)
+            refkeys.sort(key=str.lower)
         for key in refkeys:
             tbody.append(refdict[key])
         self.ref_start += i + 1                
@@ -738,7 +766,7 @@ class HtmlRfcWriter(BaseRfcWriter):
         # Grab values that haven't been inserted yet
         background_image = (self.pis['background'] and 
             "background-image:url('%s');" % self.pis['background'] or '')
-        title = self.r.find('front/title').text
+        title = self.r.find('front/title').text or ''
         docName = self.r.attrib.get('docName', '')
         description = ''
         abs_t = self.r.find('front/abstract/t')
@@ -746,7 +774,7 @@ class HtmlRfcWriter(BaseRfcWriter):
             description = abs_t.text
         keywords = self.r.findall('front/keyword')
         keyword_list = [keyword.text for keyword in keywords if keyword.text]
-        generator = "xml2rfc version %s - http://tools.ietf.org/tools/xml2rfc" % xml2rfc.__version__
+        generator = "xml2rfc version %s - https://tools.ietf.org/tools/xml2rfc" % xml2rfc.__version__
 
         # Build ISO8601 date string
         docDate = ''
@@ -787,7 +815,7 @@ class HtmlRfcWriter(BaseRfcWriter):
                  'toc_rows':        ''.join(self.buffers['toc_rows'])
                 }
         for key in subs.keys():
-            subs[key] = subs[key].encode('ascii', 'xmlcharrefreplace').decode('UTF-8')
+            subs[key] = subs[key].encode('ascii', 'xmlcharrefreplace').decode('ascii')
         self.output = self.templates['base.html'].substitute(subs)
 
     def post_process_lines(self, lines):
