@@ -4,6 +4,7 @@
 
 # Python libs
 import datetime
+
 try:
     import debug
     assert debug
@@ -11,9 +12,10 @@ except ImportError:
     pass
 
 # Local libs
-from xml2rfc.writers.base import BaseRfcWriter
+from xml2rfc.writers.base import BaseRfcWriter, default_options
 from xml2rfc.writers.raw_txt import RawTextRfcWriter
 import xml2rfc.utils
+import xml2rfc.log
 
 class PaginatedTextRfcWriter(RawTextRfcWriter):
     """ Writes to a text file, paginated with headers and footers
@@ -21,10 +23,11 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
         The page width is controlled by the *width* parameter.
     """
 
-    def __init__(self, xmlrfc, width=72, quiet=False, verbose=False, date=datetime.date.today(),
-                    omit_headers=False):
-        RawTextRfcWriter.__init__(self, xmlrfc, width=width, quiet=quiet, \
-                                  verbose=verbose, date=date)
+    def __init__(self, xmlrfc, width=72, quiet=None, options=default_options,
+                               date=datetime.date.today(), omit_headers=False):
+        if not quiet is None:
+            options.quiet = quiet
+        RawTextRfcWriter.__init__(self, xmlrfc, options=options, date=date)
         self.left_header = ''
         self.center_header = ''
         self.right_header = ''
@@ -35,6 +38,13 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
         self.paged_toc_marker = 0
         self.page_line = 0
         self.omit_headers = omit_headers
+        self.page_end_blank_lines = 2
+        # Don't permit less than this many lines of a broken paragrah at the
+        # top of a page:
+        self.widow_limit = self.get_numeric_pi('widowlimit', default=2)
+        # Don't permit less than this many lines of a broken paragrah at the
+        # bottom of a page:
+        self.orphan_limit = self.get_numeric_pi('orphanlimit', default=2)
 
     def _make_footer_and_header(self, page, final=False):
         tmp = []
@@ -89,10 +99,14 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
 
     def needLines(self, count):
         """Deal with the PI directive needLines"""
-        if int(count) < 0:
-            self._set_break_hint(1, 'break', len(self.buf))
-        else:
-            self._set_break_hint(count, 'raw', len(self.buf))
+        try:
+            if int(count) < 0:
+                self._set_break_hint(1, 'break', len(self.buf))
+            else:
+                self._set_break_hint(count, 'raw', len(self.buf))
+        except ValueError as e:
+            if not self.indexmode:
+                xml2rfc.log.warn('%s, in processing instruction needlines="%s"' % (str(e).capitalize(), count))
 
     # Here we override some methods to mark line numbers for large sections.
     # We'll store each marking as a dictionary of line_num: section_length.
@@ -144,9 +158,18 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
                                        level=level)
         # Reserve room for a blankline and some lines of section content
         # text, in order to prevent orphan headings
-        end = len(self.buf) + self.pis["sectionorphan"]
-        self._set_break_hint(end-begin, 'raw', begin)
-                                        
+        #
+        # We've now written a blank line before the heading, followed by
+        # the heading.  Next will be a blank line, followed by the section
+        # content.  We thus need (begin - end) + 1 + orphan_limit lines, but
+        # if we ask for only that, we'll end up moving the text, to the next
+        # page, without moving the section heading.  So we ask for one more
+        self.orphan_limit = self.get_numeric_pi('orphanlimit', default=self.orphan_limit)
+        orphanlines = self.get_numeric_pi('sectionorphan', default=self.orphan_limit+2)
+        end = len(self.buf)
+        needed = end - begin + orphanlines
+        self._set_break_hint(needed, 'raw', begin)
+
 
     def pre_rendering(self):
         """ Prepares the header and footer information """
@@ -250,8 +273,8 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
         for line_num, line in enumerate(self.buf):
             if line_num == self.toc_marker and self.toc_marker > 0:
                 # Don't start ToC too close to the end of the page
-                if self.page_length + 10 >= max_page_length:
-                    remainder = max_page_length - self.page_length - 2
+                if self.pis['tocpagebreak'] == 'yes' or self.page_length + 10 >= max_page_length:
+                    remainder = max_page_length - self.page_length - self.page_end_blank_lines
                     self.emit([''] * remainder)
                     self.page_break()
 
@@ -259,7 +282,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
                 toc_prev_start = len(self.output)
                 preliminary_toc = self.write_toc(paging=True)
                 for l in preliminary_toc:
-                    if self.page_length + 2 >= max_page_length:
+                    if self.page_length + self.page_end_blank_lines >= max_page_length:
                         # Store a pair of TOC pointers
                         toc_pointers.append((toc_prev_start, len(self.output)))
                         # New page
@@ -273,7 +296,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
             if line_num == self.iref_marker and self.iref_marker > 0:
                 # Don't start Index too close to the end of the page
                 if self.page_length + 10 >= max_page_length:
-                    remainder = max_page_length - self.page_length - 2
+                    remainder = max_page_length - self.page_length - self.page_end_blank_lines
                     self.emit([''] * remainder)
                     self.page_break()
 
@@ -285,7 +308,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
                 iref_prev_start = len(self.output)
                 preliminary_iref_index = self.write_iref_index()
                 for l in preliminary_iref_index:
-                    if self.page_length + 2 >= max_page_length:
+                    if self.page_length + self.page_end_blank_lines >= max_page_length:
                         # Store a pair of pointers
                         iref_pointers.append((iref_prev_start, len(self.output)))
                         # New page
@@ -300,7 +323,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
                 # If this size hint exceeds the rest of the page, or is set
                 # to -1 (a forced break), insert a break.
 
-                available = max_page_length - (self.page_length + 2)
+                available = max_page_length - (self.page_length + self.page_end_blank_lines)
                 needed, text_type = self.break_hints[line_num]
 
                 blanks = 0
@@ -319,17 +342,17 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
                     needed -= blanks
 
                 if (text_type == "break"
-                    or (text_type == "raw" and needed > available and needed < max_page_length-2 )):
+                    or (text_type == "raw" and needed > available and needed < max_page_length-self.page_end_blank_lines )):
                     lines_until_break = 0
 
                 elif (self.pis['autobreaks'] == 'yes'
                       and needed > available
-                      and needed < max_page_length-2
-                      and (needed-available < 2 or available < 2) ):
+                      and needed < max_page_length-self.page_end_blank_lines
+                      and (needed-available < self.widow_limit or available < self.orphan_limit) ):
                     lines_until_break = available
                     if available == needed - 1:
                         lines_until_break -= 1
-                    if lines_until_break < 2:
+                    if lines_until_break < self.page_end_blank_lines:
                         lines_until_break = 0
                     if lines_until_break > 0 and blanks > 0:
                         # put back that blank line since we are going to emit it
@@ -337,12 +360,12 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
 
             if lines_until_break == 0:
                 # Insert break
-                remainder = max_page_length - self.page_length - 2
+                remainder = max_page_length - self.page_length - self.page_end_blank_lines
                 self.emit([''] * remainder)
             if not self.IsFormatting(line):
                 lines_until_break -= 1
 
-            if self.page_length + 2 >= max_page_length:
+            if self.page_length + self.page_end_blank_lines >= max_page_length:
                 # New page
                 self.page_break()
 
@@ -363,7 +386,7 @@ class PaginatedTextRfcWriter(RawTextRfcWriter):
 
         # Write final footer
         if self.page_length > 1:
-            remainder = max_page_length - self.page_length - 2
+            remainder = max_page_length - self.page_length - self.page_end_blank_lines
             self.emit([''] * remainder)
         self.page_break(final=True)
         
