@@ -4,9 +4,10 @@
 
 # Python libs
 import textwrap
-import lxml
 import datetime
+import lxml
 import re
+
 try:
     import debug
     assert debug
@@ -14,7 +15,7 @@ except ImportError:
     pass
 
 # Local lib
-from xml2rfc.writers.base import BaseRfcWriter
+from xml2rfc.writers.base import BaseRfcWriter, default_options
 import xml2rfc.utils
 
 
@@ -24,8 +25,10 @@ class RawTextRfcWriter(BaseRfcWriter):
         The page width is controlled by the *width* parameter.
     """
 
-    def __init__(self, xmlrfc, width=72, margin=3, quiet=False, verbose=False, date=datetime.date.today()):
-        BaseRfcWriter.__init__(self, xmlrfc, quiet=quiet, verbose=verbose, date=date)
+    def __init__(self, xmlrfc, width=72, margin=3, quiet=None, options=default_options, date=datetime.date.today()):
+        if not quiet is None:
+            options.quiet = quiet
+        BaseRfcWriter.__init__(self, xmlrfc, options=options, date=date)
         # Document processing data
         self.width = width      # Page width
         self.margin = margin    # Page margin
@@ -36,7 +39,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.list_counters = {} # Maintain counters for 'format' type lists
         self.edit_counter = 0   # Counter for edit marks
         # Set this to False to permit utf+8 output:
-        self.ascii = True       # Enable ascii flag
+        self.ascii = not options.utf8       # Enable ascii flag
         self.cref_counter = 0   # Counter for cref anchors
         self.cref_list = []
 
@@ -163,7 +166,7 @@ class RawTextRfcWriter(BaseRfcWriter):
             # Check for optional hangIndent
             try:
                 hangIndent = list.attrib.get('hangIndent', None)
-                if hangIndent:
+                if hangIndent != None:
                     hangIndent = int(hangIndent)
             except (ValueError, TypeError):
                 xml2rfc.log.error("hangIndent value '%s' is not an integer" % hangIndent)
@@ -188,14 +191,16 @@ class RawTextRfcWriter(BaseRfcWriter):
         for element in list:
             # Check for PI
             if element.tag is lxml.etree.PI:
-                self.xmlrfc.parse_pi(element)
+                pidict = self.parse_pi(element)
+                if pidict and "needLines" in pidict:
+                    self.needLines(pidict["needLines"])
             elif element.tag == 't':
                 # Disable linebreak if subcompact=yes AND not first list element
                 leading_blankline = True
-                if t_count > 0 and self.pis['subcompact'] == 'yes':
+                if t_count > 0 and element.pis['subcompact'] == 'yes':
                     leading_blankline = False
                 if style == 'symbols':
-                    bullet = self.pis['text-list-symbols'][level % len(self.pis['text-list-symbols'])]
+                    bullet = element.pis['text-list-symbols'][level % len(element.pis['text-list-symbols'])]
                     bullet += '  '
                 elif style == 'numbers':
                     bullet = self._format_counter("%d.", t_count+1, listlength)
@@ -204,7 +209,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                     bullet = self._format_counter(letter_style, t_count+1, listlength)
                 elif style == 'hanging':
                     bullet = element.attrib.get('hangText', '')
-                    if not hangIndent:
+                    if hangIndent is None:
                         hangIndent = 3
                     if len(bullet) < hangIndent:
                         # Insert whitespace up to hangIndent
@@ -214,7 +219,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                         bullet += '  '
                     # Add an extra space in front of colon if colonspace enabled
                     if bullet.endswith(':') and \
-                    self.pis['colonspace'] == 'yes':
+                    element.pis['colonspace'] == 'yes':
                         bullet+= ' '
                     if element.text and len(bullet) > self.width//2:
                         # extra check of remaining space if the bullet is
@@ -289,7 +294,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                 indent = 3
                 sub_indent = indent + len(bullet)
                 pagestr = '%4s' % item.page
-                lines = textwrap.wrap(bullet + (item.title if item.title else ""),
+                lines = textwrap.wrap(bullet + (item.title.strip() if item.title else ""),
                                       self.width - len(pagestr),
                                       initial_indent=' ' * indent,
                                       subsequent_indent=' ' * sub_indent)
@@ -373,11 +378,13 @@ class RawTextRfcWriter(BaseRfcWriter):
         if not item:
             target_text = '[' + target + ']'
         elif format == 'none':
-            return xref.text.rstrip()
+            if xref.text:
+                return xref.text.rstrip()
+            return ''
         elif format == 'counter':
             target_text = item.counter
         elif format == 'title':
-            target_text = item.title
+            target_text = item.title.strip()
         else: #Default
             target_text = item.autoName
         target_text = re.sub("-", u"\u2011", target_text) # switch to non-breaking hyphens
@@ -423,9 +430,6 @@ class RawTextRfcWriter(BaseRfcWriter):
         """
         line = ['']
         for i, element in enumerate(elements):
-            # Check for a PI first
-            if element.tag is lxml.etree.PI:
-                self.xmlrfc.parse_pi(element)
             if element.tag not in self.inline_tags:
                 # Not an inline element, exit
                 return ''.join(line), elements[i:]
@@ -433,17 +437,18 @@ class RawTextRfcWriter(BaseRfcWriter):
             if element.tag == 'xref':
                 line.append(self._expand_xref(element))
             elif element.tag == 'eref':
-                if element.text:
-                    line.append(element.text + ' ')
-                self.eref_count += 1
-                if element.text != element.attrib['target']:
-                    line.append('[' + str(self.eref_count) + ']')
+                if element.text and len(element.text)>0:
+                    line.append(element.text.strip())
+                    self.eref_count += 1
+                    line.append(' [' + str(self.eref_count) + ']')
                     if self.indexmode:
                         self.eref_list.append([self.eref_count, element])
+                else:
+                    line.append('<' + element.attrib['target'].strip() + '>')
             elif element.tag == 'iref':
                 self._add_iref_to_index(element)
             elif element.tag == 'cref' and \
-                self.pis['comments'] == 'yes':                
+                element.pis['comments'] == 'yes':                
                 # Render if processing instruction is enabled
                 self.cref_counter += 1
                 anchor = element.attrib.get('anchor', None)
@@ -451,7 +456,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                     anchor = 'CREF' + str(self.cref_counter)
                     element.attrib['anchor'] = anchor
                 self._indexCref(self.cref_counter, anchor)
-                if self.pis['inline'] == 'yes':
+                if element.pis['inline'] == 'yes':
                     if anchor:
                         anchor = anchor + ': '
                     source = element.attrib.get('source', "")
@@ -464,7 +469,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                     self.cref_list.append(element)
             elif element.tag == 'spanx':
                 style = element.attrib.get('style', 'emph')
-                edgechar = '?'
+                edgechar = '_'  # default to emph because the spanx element exists
                 if style == 'emph':
                     edgechar = '_'
                 elif style == 'strong':
@@ -738,16 +743,19 @@ class RawTextRfcWriter(BaseRfcWriter):
                 refstring.append(', ')
             title = ref.find('front/title')
             if title is not None and title.text:
-                refstring.append('"' + title.text.strip() + '", ')
+                if ref.attrib.get("quote-title", "true") == "true": # attribute default value: yes
+                    refstring.append('"' + title.text.strip() + '"')
+                else:
+                    refstring.append(title.text.strip())
             else:
                 xml2rfc.log.warn('No title specified in reference',
                                  ref.attrib.get('anchor', ''))
             for seriesInfo in ref.findall('seriesInfo'):
                 if seriesInfo.attrib['name'] == "Internet-Draft":
-                    refstring.append(seriesInfo.attrib['value'] + ' (work in progress), ')
+                    refstring.append(', '+seriesInfo.attrib['value'] + ' (work in progress)')
                 else:
-                    refstring.append(seriesInfo.attrib['name'] + ' ' +
-                                     seriesInfo.attrib['value'] + ', ')
+                    refstring.append(', '+seriesInfo.attrib['name'] + u'\u00A0' +
+                                     seriesInfo.attrib['value'].replace('/', '/' + u'\u200B'))
             date = ref.find('front/date')
             if date is not None:
                 month = date.attrib.get('month', '')
@@ -755,13 +763,11 @@ class RawTextRfcWriter(BaseRfcWriter):
                 if month or year:
                     if month:
                         month += ' '
-                    refstring.append(month + year)
+                    refstring.append(', '+month + year)
             # Target?
             target = ref.attrib.get('target')
             if target:
-                if not refstring[-1].endswith(', '):
-                    refstring.append(', ')
-                refstring.append('<' + target + '>')
+                refstring.append(', <' + target + '>')
             refstring.append('.')
             annotation = ref.find('annotation')
             # Use anchor or num depending on PI
@@ -785,7 +791,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         # numeric order, and if we sort, they will be sorted alphabetically,
         # rather than numerically ... ( i.e., [10], [11], ... [19], [1], ... )
         if self.pis['sortrefs'] == 'yes' and self.pis['symrefs'] == 'yes' :
-            refkeys.sort()
+            refkeys.sort(key=str.lower)
         # Hard coded indentation amount
         refindent = 11
         for key in refkeys:
@@ -848,7 +854,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         for i, cell in enumerate(table.findall('c')):
             if i % num_columns == 0:
                 # Insert blank row if PI 'compact' is 'no'
-                if self.pis["compact"] == "no":
+                if table.pis["compact"] == "no":
                     if (style == 'none' and row == 0) or (style in ['headers', 'full'] and row != 0):
                         row += 1
                         matrix.append(['']*num_columns)
@@ -1146,6 +1152,8 @@ class RawTextRfcWriter(BaseRfcWriter):
                    plus.append(a-b)
                target, ttcol_widths = self.expand_cols(longest_words, target, plus)
 
+        # Ensure we don't have any zero-width columns; it breaks textwrap
+        ttcol_widths = [ k or 1 for k in ttcol_widths ] 
         # Now construct the cells using textwrap against ttcol_widths
         cell_lines = [
             [ textwrap.wrap(cell, ttcol_widths[j]) or [''] for j, cell in enumerate(matrix[i]) ]
@@ -1414,6 +1422,7 @@ class RawTextRfcWriter(BaseRfcWriter):
                 # line = line.replace(u"\u8209", '-')
                 line = line.replace(u'\u00A0', u' ')
                 line = line.replace(u'\u2011', u'-')
+                line = line.replace(u'\u200B', u'')
             output.append(line);
         return output
 
